import os
import time
import sys

from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.table import Table
from astropy.time import Time
import astropy.units as u
import numpy as np
from pandas import Series

import ssos.filt as filt
from ssos.opt import extract_cutouts
from ssos.opt import compute_aperture_magnitudes
from ssos.opt import crossmatch_skybot
from ssos.utils import create_target_dir
from ssos.utils import init_argparse
from ssos.utils import init_logger


FILTER_STEPS = {
        'FILTER_DETEC':        filt.detections,
        'FILTER_PM':           filt.proper_motion,
        'FILTER_PIXEL':        filt.pixel,
        'FILTER_MOTION':       filt.linear_motion,
        'FILTER_TRAIL':        filt.constant_trail,
        'FILTER_T_DIST':       filt.trail_distribution,
        'FILTER_STAR_REGIONS': filt.star_catalog
        }


ANALYSIS_STEPS = {
    'CROSSMATCH_SKYBOT': crossmatch_skybot,
    'EXTRACT_CUTOUTS':   extract_cutouts,
    'FIXED_APER_MAGS':   compute_aperture_magnitudes
    }


class Pipeline:
    ''' Class to handle the pipeline execution '''

    def __init__(self):

        # Time pipeline execution
        self.start_time = time.time()

        # Preparations: Handling of command line arguments
        self.args = init_argparse()
        self.target_dir, self.paths = create_target_dir(self.args)
        self.log, self.log_file = init_logger(self.args, self.paths['logs'])
        self.log.info('\n\t--- SSO Recovery Pipeline ---\t\t--- {} ---\n\n'
                      .format(time.strftime("%H:%M:%S %Y/%m/%d")))

        # Reading and checking the settings
        self.settings = self._set_settings()
        self.settings = self._check_settings(self.settings)

        # Assert that images are found and contain the required header keywords
        self.images = [os.path.join(self.paths['images'], image) for image in
                       os.listdir(self.paths['images']) if image.endswith('.fits')]
        assert len(self.images) > 0, 'No images found in %s! Assure that they have a .fits \
                                      extension' % self.paths['images']
        self._print_field_info()

        self.steps = [param for param in FILTER_STEPS.keys() if self.settings[param]]
        self.analysis_steps = [param for param in ANALYSIS_STEPS.keys() if self.settings[param]]
        self.added_proper_motion = False
        self.added_trail_morphology = False

        # Little hack for fixed apertuer magnitudes "without" cutout extraction
        if not self.settings['EXTRACT_CUTOUTS'] and self.settings['FIXED_APER_MAGS']:
            self.log.debug('FIXED_APER_MAGS is True but EXTRACT_CUTOUTS is False. Will\
                            delete cutouts upon finishing.')
            self.paths['cutouts'] = self.paths['tmp']
            self.analysis_steps[0:0] = ['EXTRACT_CUTOUTS']


    def _set_settings(self):
        '''
        Set pipeline settings

        input
        ------
        args - command line arguments parser
        log - logging instance

        return
        ------
        settings - dict, dictionary containing PARAMETER: VALUE pairs
        '''

        settings = {}

        try:
            if self.args.set_up:
                set_up = open(self.args.set_up[0], 'r')
            else:
                set_up = open('pipeline_settings.ssos', 'r')

        except IOError:
            self.log.debug('No config file provided in CWD or via -c flag.\
                            Using default settings.\n')
            try:
                set_up = open(os.path.join(os.path.dirname(__file__),
                                           'pipeline_settings.ssos'))
            except IOError:
                raise PipelineSettingsException('No configuration file provided\
                       and not found in package directory.')

        with set_up:
            for line in set_up:
                if line == '\n' or line[0] == '#':
                    continue

                parameter, value = line.split()[:2]
                settings[parameter] = value

        # Override values from command line
        for arg in vars(self.args):
            if arg in settings.keys() and vars(self.args)[arg] != None:
                settings[arg] = vars(self.args)[arg]

        # Expand $HOME to home directory
        settings = {k: v.replace('$HOME', os.path.expanduser('~')) for k, v in settings.items()}

        # Filter setting finished. Write them to the logfile
        with open(os.path.join(self.paths['logs'], self.log_file), 'a') as logfile:

            for setting, value in settings.items():
                line = setting.ljust(20) + str(value) + '\n'
                self.log.debug(line)
                logfile.write(line)

            logfile.write('\n')
            self.log.debug('\n')

        return settings


    def _check_settings(self, settings):
        '''
        Converts parameter values to expected formats

        input
        -----
        settings - dict, dictionary containing PARAMETER: VALUE pairs

        return
        -----
        settings - dict, dictionary containing PARAMETER: VALUE pairs
        '''

        # Check image header keywords
        with fits.open(self.images[0]) as exposure:

            header = exposure[self.settings['SCI_EXTENSION'][0]].header

            for kw in ['RA', 'DEC', 'OBJECT', 'DATE-OBS', 'FILTER', 'EXPTIME']:
                try:
                    _ = header[self.settings[kw]]

                except KeyError:
                    raise PipelineSettingsException('Could not find keyword %s in FITS header.\
                                                     Is the SCI_EXTENSION correct?' % kw)


        # Check that config files exist
        for file in ['SEX_CONFIG', 'SEX_PARAMS', 'SEX_NNW', 'SEX_FILTER',
                     'SCAMP_CONFIG', 'SWARP_CONFIG', 'HYGCAT']:
            if not os.path.isfile(settings[file]):
                raise PipelineSettingsException('Could not find %s in %s' %
                                                (file, settings[file]) )

        # Check if weight images are provided and exist
        if not settings['WEIGHT_IMAGES'].upper() == 'FALSE':
            if not os.path.isdir(settings['WEIGHT_IMAGES']):
                raise PipelineSettingsException('Could not find weight images directory %s.' %
                                                 settings['WEIGHT_IMAGES'])
        else:
            settings['WEIGHT_IMAGES'] = False

        # Unpack provided DETECTIONS and SCI extension(s) into list
        for param in ['SCI_EXTENSION', 'DETECTIONS']:
            try:
                settings[param] = [int(character) for character in settings[param].split(',')]
            except ValueError:
                raise PipelineSettingsException('%s value invalid' % param)

        # Convert filter strings to booleans
        for param in ['FILTER_DETEC', 'FILTER_PM', 'FILTER_PIXEL', 'FILTER_MOTION',
                      'IDENTIFY_OUTLIER', 'FILTER_TRAIL', 'FILTER_T_DIST',
                      'FILTER_STAR_REGIONS', 'CROSSMATCH_SKYBOT', 'EXTRACT_CUTOUTS',
                      'FIXED_APER_MAGS']:

            if settings[param].upper() == 'TRUE':
                settings[param] = True
            elif settings[param].upper() == 'FALSE':
                settings[param] = False
            else:
                raise PipelineSettingsException('Could not evaluate %s value, has to be True or\
                                                 False' % param)

        # Convert numeric values to float
        for param in ['PM_LOW', 'PM_UP', 'PM_SNR', 'DELTA_PIXEL', 'OUTLIER_THRESHOLD',
                      'R_SQU_M', 'RATIO', 'SIGMA', 'DISTANCE', 'CROSSMATCH_RADIUS',
                      'CUTOUT_SIZE']:

            try:
                settings[param] = float(settings[param])
            except ValueError:
                raise PipelineSettingsException('Could not convert %s value to float'
                                                % param)
            if param == ['R_SQU_M']:
                try:
                    assert 0 <= settings[param] <= 1
                except AssertionError:
                    raise PipelineSettingsException('The %s parameter has to be in range [0 - 1]'
                                                     % param)
        return settings


    def _print_field_info(self):
        ''' Prints RA, DEC, and OBJECT keywords to log '''
        ra = self.settings['RA']
        dec = self.settings['DEC']
        object_ = self.settings['OBJECT']

        ecli_lat = SkyCoord(ra, dec, frame='icrs', unit='deg').barycentrictrueecliptic.lat.deg

        self.log.info('\t|\t'.join(['%i Exposures' % len(self.images),
                                    '%s' % object_,
                                    '%.2fdeg Ecliptic Latitude\n\n' % ecli_lat]))


    def _run_SExtractor_on_single_image(self, image, extension):
        '''
        Run SExtractor on individual images.

        input
        ------
        image - str, absolute path to image file
        extension - int, index of FITS file extension to be SExtracted

        return
        ------
        cat, str - absolute path to SExtractor output catalog
        '''

        # Name of output catalog
        cat = os.path.join(self.paths['cats'],
                           os.path.splitext(os.path.basename(image))[0]) + '_%i.cat' % extension

        if not self.args.sex and os.path.isfile(cat):
            self.log.debug('SExtractor catalog %s already exists! Skipping this sextraction..\n' % cat)
            return cat

        sex_args = {

            'file': image + '[%i]' % extension,
            'config': self.settings['SEX_CONFIG'],
            'overwrite_params': { # Arguments to initialize Astromatic class
                    'CATALOG_NAME': cat,
                    'PARAMETERS_NAME': self.settings['SEX_PARAMS'],
                    'FILTER_NAME': self.settings['SEX_FILTER'],
                    'STARNNW_NAME': self.settings['SEX_NNW'],
            },
        }

        if self.settings['WEIGHT_IMAGES']:
            sex_args['overwrite_params']['WEIGHT_IMAGE'] = os.path.join(self.settings['WEIGHT_IMAGES'],
                                                           os.path.basename(image).replace(
                                                           '.fits', '_%i.weight' % extension))

        if self.log.level <= 10:  # if we're at DEBUG log level, print SExtractor output
            sex_args['overwrite_params']['VERBOSE_TYPE'] = 'NORMAL'

        # ------
        # Exectue SExtractor
        cmd = ' '.join(['sex', sex_args['file'], '-c', sex_args['config']])
        for param, value in sex_args['overwrite_params'].items():
            cmd += ' '.join([' -' + param, value])

        self.log.debug('\nExecuting SExtractor command:\n%s\n\n' % cmd)
        os.system(cmd)

        # ------
        # Check if MJD-OBS keyword is in header. Else, create ahead file for SCAMP
        with fits.open(image) as exposure:

            try:
                mjd_obs = exposure[extension].header['MJD-OBS']

            except KeyError:
                # Make .ahead file with the EPOCH in MJD for SCAMP
                # Following http://www.astromatic.net/forum/showthread.php?tid=501
                with fits.open(image) as file:
                    date_obs = file[extension].header[self.settings['DATE-OBS']]

                mjd = Time(date_obs, format='isot').mjd

                with open(os.path.splitext(cat)[0] + '.ahead', 'w+') as file:
                    file.write('MJD-OBS = %.6f' % mjd)

        return cat


    def run_SExtractor(self):
        ''' Wrapper for SExtractor run.  Adds progress bar.'''

        self.SExtractor_catalogues = []

        # Add progress bar
        if not self.args.quiet:
            sys.stdout.write('[%s]' % (' ' * len(self.images)))
            sys.stdout.flush()
            sys.stdout.write('\b' * (len(self.images) + 1))

        for image in self.images:

            for extension in self.settings['SCI_EXTENSION']:

                cat = self._run_SExtractor_on_single_image(image, extension)
                self.SExtractor_catalogues.append(cat)

            if not self.args.quiet:
                sys.stdout.write('-')
                sys.stdout.flush()


    def run_SCAMP(self):
        ''' Run SCAMP on SExtractor catalogs '''

        # ------
        # Set-up SCAMP run
        self.merged_cat = os.path.join(self.paths['cats'], 'merged_1.cat')
        self.full_cat = os.path.join(self.paths['cats'], 'full_1.cat')

        scamp_args = {
            'files':  self.SExtractor_catalogues, # catalogs to SCAMP
            'config': self.settings['SCAMP_CONFIG'],

            'overwrite_params': {
                'MERGEDOUTCAT_NAME': self.merged_cat[:-6] + '.cat', # SCAMP appends '_1' on its own
                'FULLOUTCAT_NAME':   self.full_cat[:-6] + '.cat',
                },
            }

        if not self.args.scamp and os.path.isfile(self.merged_cat) and os.path.isfile(self.full_cat):
            self.log.debug('SCAMP catalogs %s and %s already exist!\
                            Skipping SCAMP run..\n' % ( self.merged_cat, self.full_cat))

        else:

            # ------
            # Exectue SCAMP
            sorted_catalogs = ' '.join([cat for cat in sorted(scamp_args['files'])])
            cmd = ' '.join(['scamp', sorted_catalogs, '-c', scamp_args['config']])
            for param, value in scamp_args['overwrite_params'].items():
                cmd += ' '.join([' -' + param, value])

            self.log.debug('Executing SCAMP command:\n%s\n' % cmd)
            if self.log.level > 10:  # Shown SCAMP output only when debugging
                    cmd += ' >/dev/null 2>&1'
            os.system(cmd)

        # Add the SCAMP catalog numbers to the SExtactor catalogs dictionary
        self.SExtractor_catalogues = {i: filename for i, filename in
                                                enumerate(sorted(self.SExtractor_catalogues), 1)}

        # Create the full catalog as pipeline property
        with fits.open(self.full_cat) as full:
            data = Table(full[2].data)

        self.sources = data.to_pandas()

        # Ensure that EPOCHs are in order
        self.sources = self.sources.sort_values(['SOURCE_NUMBER', 'EPOCH'],
                                                ascending=[False, False])
        self.sources = self.sources.reset_index()
        # Catalogue number 0 signals bad detections in SExtractor
        self.sources = self.sources[self.sources['CATALOG_NUMBER'] != 0]
        self.log.info('Done.\n')


    def number_of_sources(self):
        return len(set(self.sources.SOURCE_NUMBER))


    def check_remaining(func):
        ''' Wrapper to check if source candidates remain after filter step'''
        def wrapped_func(*args, **kwargs):

            result = func(*args, **kwargs)

            try:
                assert args[0].number_of_sources() > 0
            except AssertionError:
                raise NoSourcesRemainingException('\nNo source candidates left.'
                                                   ' Stopping pipeline.\n')
                sys.exit()

            return result
        return wrapped_func


    @check_remaining
    def execute_filter(self, step_name):
        '''
        Calls the proper filter function

        input
        ------
        step_name - str, name of filter step

        return
        ------
        sources - pd.DataFrame, remaining source candidates
        '''

        # The proper motion filter has to access the merged catalogue first
        if step_name == 'FILTER_PM':
            self.add_proper_motion()

        # The trail filters have to access the SExtractor catalogues first
        if step_name in ['FILTER_PIXEL', 'FILTER_TRAIL', 'FILTER_T_DIST']:
            self.add_trail_morphology()


        self.sources = FILTER_STEPS[step_name](self.sources, self.settings)


    def add_proper_motion(self):
        '''
        Adds proper motion values from merged to full catalog.  Has to be called
        either before filtering the proper motion or at the end of the pipeline.
        '''

        with fits.open(self.merged_cat) as merged:
            data = Table(merged[2].data)

        # Pandas cannot handle multi-dimensional columns which SCAMP outputs
        # for magnitudes in different filters
        cols = tuple(name for name in data.colnames if len(data[name].shape) <= 1)
        data = data[cols]

        sources_merged = data.to_pandas()
        sources_merged = sources_merged[
                         sources_merged.SOURCE_NUMBER.isin(self.sources.SOURCE_NUMBER)
                         ]

        for column in ['PMALPHA_J2000', 'PMDELTA_J2000', 'PMALPHAERR_J2000', 'PMDELTAERR_J2000']:
            sources_merged[column] = sources_merged[column].apply(lambda x: x / 8.75e6)

        # Add proper motion column.  Factor 8.75e6 converts to "/h
        sources_merged['PM'] = sources_merged.apply(lambda x:
                               np.sqrt(x['PMALPHA_J2000']**2 + x['PMDELTA_J2000']**2) / 8.75e6,
                               axis=1)
        sources_merged['PMERR'] = sources_merged.apply(lambda x:
                               np.sqrt(x['PMALPHAERR_J2000']**2 + x['PMDELTAERR_J2000']**2) / 8.75e6,
                               axis=1)

        # Add proper motions column from merged catalog to source database
        self.sources = self.sources.merge(sources_merged[['SOURCE_NUMBER', 'PM', 'PMERR',
                                                          'PMALPHA_J2000', 'PMALPHAERR_J2000',
                                                          'PMDELTA_J2000', 'PMDELTAERR_J2000']],
                                                          on='SOURCE_NUMBER')
        # Add flag to dataframe
        self.sources['FLAGS_SSOS'] = 0

        self.sources_merged = sources_merged
        self.added_proper_motion = True


    def add_trail_morphology(self):
        '''
        Adds the trail morphology parameters from the SExtractor catalogues to
        the source database.  Has to be called either before filtering on trail
        morphology or at the end of the pipeline.
        '''

        sources = self.sources

        for catalog_number, group in sources.groupby('CATALOG_NUMBER'):

            with fits.open(self.SExtractor_catalogues[catalog_number]) as cat:
                xwin_img = Series(cat[2].data.field('XWIN_IMAGE')) # SExtractor X pixel coordinates

                for index, source in group.iterrows():
                    index_in_sex_cat = xwin_img[xwin_img == source['X_IMAGE']].index[0]

                    for prop in ['AWIN_IMAGE', 'BWIN_IMAGE', 'THETAWIN_IMAGE',
                                 'XWIN_IMAGE', 'YWIN_IMAGE',
                                 'ERRAWIN_IMAGE', 'ERRBWIN_IMAGE', 'ERRTHETAWIN_IMAGE']:
                        sources.loc[index, prop] = cat[2].data.field(prop)[index_in_sex_cat]

                    # Replace SCAMP magnitudes with the SExtractor ones
                    for prop in ['MAG', 'MAGERR', 'FLUX', 'FLUXERR']:
                        sources.loc[index, prop] = cat[2].data.field(prop + '_AUTO')[index_in_sex_cat]

                # Add the image filename and science extension
                image_filename = os.path.splitext(
                                 os.path.basename(self.SExtractor_catalogues[catalog_number]))[0][:-2]\
                                 + '.fits'  # [:-3] removes the extension suffix

                sci_ext = int(os.path.splitext(self.SExtractor_catalogues[catalog_number])[0][-1])

                sources.loc[group.index, 'FILENAME_EXP'] = image_filename
                sources.loc[group.index, 'SCI_EXTENSION'] = sci_ext

                # Add exposure keywords
                with fits.open(os.path.join(self.paths['images'], image_filename)) as exposure:

                    sources.loc[group.index, 'OBJECT_EXP']   = exposure[sci_ext].header[self.settings['OBJECT']]
                    sources.loc[group.index, 'DATE-OBS_EXP'] = exposure[sci_ext].header[self.settings['DATE-OBS']]
                    sources.loc[group.index, 'FILTER_EXP']   = exposure[sci_ext].header[self.settings['FILTER']]
                    sources.loc[group.index, 'EXPTIME_EXP']  = exposure[sci_ext].header[self.settings['EXPTIME']]

                    sources.loc[group.index, 'RA_EXP']   = exposure[sci_ext].header[self.settings['RA']]
                    sources.loc[group.index, 'DEC_EXP']  = exposure[sci_ext].header[self.settings['DEC']]

        sources['MID_EXP_MJD'] = sources.apply(lambda x: (Time(x['DATE-OBS_EXP'], format='isot') +
                                               float(x['EXPTIME_EXP']) / 2 * u.second).mjd, axis=1)

        self.sources = sources
        self.added_trail_morphology = True


    def execute_analysis(self, step_name):
        '''
        Calls the proper analysis function

        input
        ------
        step_name - str, name of analysis step
        '''
        self.sources = ANALYSIS_STEPS[step_name](self.sources, self.settings, self.log,
                                      self.paths, self.args)


    def save_and_cleanup(self):
        ''' Clean up the final database.  Rename columns, remove unnecessary columns. '''

        self.sources.drop(columns=['EXTENSION', 'ASTR_INSTRUM', 'PHOT_INSTRUM'], inplace=True)

        self.sources.rename(index=str, columns={'ALPHA_J2000': 'RA', 'DELTA_J2000': 'DEC', 'PMALPHA_J2000': 'PMRA',
                                                'PMALPHAERR_J2000': 'PMRA_ERR', 'PMDELTA_J2000': 'PMDEC',
                                                'PMDELTAERR_J2000': 'PMDEC_ERR', 'SKYBOT_PMALPHA': 'SKYBOT_PMRA',
                                                'SKYBOT_PMDELTA': 'SKYBOT_PMDEC', 'SKYBOT_ALPHA': 'SKYBOT_RA',
                                                'SKYBOT_DELTA': 'SKYBOT_DEC'}, inplace=True)

        # Rearrange columns
        ordered_columns = ['SOURCE_NUMBER', 'CATALOG_NUMBER', 'RA', 'DEC', 'EPOCH', 'MAG', 'MAGERR', 'FLUX', 'FLUXERR',
                           'PM', 'PMERR', 'PMRA',  'PMRA_ERR', 'PMDEC', 'PMDEC_ERR', 'MID_EXP_MJD', 'DATE-OBS_EXP',
                           'EXPTIME_EXP', 'OBJECT_EXP', 'FILTER_EXP', 'RA_EXP', 'DEC_EXP', 'FILENAME_EXP',
                           'SCI_EXTENSION', 'XWIN_IMAGE', 'YWIN_IMAGE', 'AWIN_IMAGE', 'ERRAWIN_IMAGE', 'BWIN_IMAGE',
                           'ERRBWIN_IMAGE', 'THETAWIN_IMAGE', 'ERRTHETAWIN_IMAGE', 'ERRA_WORLD', 'ERRB_WORLD',
                           'ERRTHETA_WORLD', 'FLAGS_EXTRACTION', 'FLAGS_SCAMP', 'FLAGS_IMA', 'FLAGS_SSOS']

        if self.settings['FIXED_APER_MAGS']:
            ordered_columns[7:7] = ['MAG_CI', 'MAGERR_CI', 'FLUX_CI', 'FLUXERR_CI']

        if self.settings['CROSSMATCH_SKYBOT']:
            ordered_columns[22:22] = ['SKYBOT_NUMBER', 'SKYBOT_NAME', 'SKYBOT_CLASS', 'SKYBOT_MAG',
                                      'SKYBOT_RA', 'SKYBOT_DEC', 'SKYBOT_PMRA', 'SKYBOT_PMDEC']

        self.sources = self.sources[ordered_columns]

        # Save final database
        output = Table.from_pandas(self.sources)
        output_filename = os.path.join(self.paths['cats'], 'ssos.csv')
        output.write(output_filename, format='csv', overwrite=True)

        os.system('rm -r %s' % self.paths['tmp'])
        self.run_time = time.time() - self.start_time


class PipelineSettingsException(Exception):
    pass


class NoSourcesRemainingException(Exception):
    pass













import argparse
import logging
import os
import sys
import time


def init_argparse():
    '''
    Executes the setup of the arguments parser for the command-line api

    return
    ------
    args - command line argument parser
    '''

    parser = argparse.ArgumentParser(description='Pipeline to search for Solar System\
                                                  objects in wide-field imaging surveys')

    parser.add_argument('fields', nargs='+', help='Path to directory of field to be searched')

    parser.add_argument('-c', '--config', action='store', dest='set_up', nargs=1,
                        help='Path to configuration file')

    parser.add_argument('-t', '--target', action='store', dest='target', nargs=1,
                        help='Target directory to save fits files. If no target given, writing to CWD')

    parser.add_argument('-l', '--log', action='store', dest='log', nargs=1,
                        help='Set the logging level. Valid arguments are DEBUG, INFO, WARNING, ERROR,\
                        CRITICAl.', default=['INFO'])

    parser.add_argument('-q', '--quiet', action='store_true',
                        help='Suppress logging to console')

    parser.add_argument('--sex', action='store_true',
                        help='Force SExtractor runs')

    parser.add_argument('--scamp', action='store_true',
                        help='Force SCAMP runs')

    parser.add_argument('--swarp', action='store_true',
                        help='Force SWARP runs')

    parser.add_argument('--skybot', action='store_true',
                        help='Force SkyBoT query')

    group = parser.add_argument_group('Filter Settings')

    for prop in ['FILTER_DETEC', 'FILTER_PM', 'FILTER_PIXEL', 'FILTER_MOTION', 'IDENTIFY_OUTLIERS',
                 'FILTER_TRAIL', 'FILTER_T_DIST', 'FILTER_STAR_REGIONS', 'CROSSMATCH_SKYBOT',
                 'EXTRACT_CUTOUTS', 'FIXED_APER_MAGS']:

        group.add_argument('-'+prop, metavar='bool', action='store',
                           help='Override ' + prop + ' setting. Must be True or False.',
                           choices=['True', 'False'])


    for val in ['SCI_EXTENSION', 'WEIGHT_IMAGES', 'RA', 'DEC', 'OBJECT', 'DATE_OBS', 'FILTER', 'EXPTIME',
                'SEX_CONFIG', 'SEX_PARAMS', 'SEX_FILTER', 'SEX_NNW', 'SCAMP_CONFIG', 'SWARP_CONFIG',
                'DETECTIONS', 'PM_LOW', 'PM_UP', 'PM_SNR', 'DELTA_PIXEL', 'OUTLIER_THRESHOLD', 'R_SQU_M',
                'R_SQU_T', 'SIGMA', 'DISTANCE', 'HYGCAT', 'CROSSMATCH_RADIUS', 'CUTOUT_SIZE',
                'REFERENCE_FILTER', 'OBSERVATORY_CODE', 'FOV_DIMENSIONS']:

        group.add_argument('-'+val, metavar='value', action='store',
                           help='Override ' + val + ' setting.')

    args = parser.parse_args()

    return args


def init_logger(args, log_dir):
    '''
    Initiate logging. By default, the log is written to file.
    The filename consists of the string 'sso_' followed by the current date and time.

    input
    ------
    args - command line argument parser
    log_dir - str, absolute path to log directory

    return
    ------
    log - logger instance
    log_file - str, absolute path to log file
    '''

    # See if logging level was set via argument. Else, default to INFO
    numeric_level = getattr(logging, args.log[0].upper(), None)

    if not isinstance(numeric_level, int):
        numeric_level = 20

    log = logging.getLogger(__name__)
    log.setLevel(numeric_level)

    log_file = 'sso_{}.log'.format(time.strftime('%Y%m%d%H%M%S'))

    file_log = logging.FileHandler(os.path.join(log_dir, log_file))
    file_log.setLevel(numeric_level)
    file_log.terminator = '' # suppress newline characters
    log.addHandler(file_log)

    # If the quiet flag is set, logs to console are supressed
    if not args.quiet:
        stdout_log = logging.StreamHandler()
        stdout_log.setLevel(numeric_level)
        stdout_log.terminator = '' # suppress newline characters
        log.addHandler(stdout_log)

    def log_excepthook(execption_type, exception_value, traceback, logger=log):
        ''' Function to pipe tracebacks into the logfile '''
        log.error('Exception occurred during analysis: \n',
                  exc_info=(execption_type, exception_value, traceback))
        log.error('\n')

    sys.excepthook = log_excepthook
    return log, log_file


def create_target_dir(args):
    '''
    Creates the target directory file structure

    input
    ------
    args - command line argument parser

    return
    ------
    target_dir - str, absolute path to target directory
    paths - dict, dictionary containing the absolute paths of the subdirectories
    '''

    if not args.target:
        target_dir = os.getcwd()
    else:
        target_dir = os.path.abspath(args.target[0])

    paths = {
        'cats':    os.path.join(target_dir, 'cats'),
        'cutouts': os.path.join(target_dir, 'cutouts'),
        'logs':     os.path.join(target_dir, 'logs'),
        'images':  os.path.abspath(args.fields[0]),
        'skybot':  os.path.join(target_dir, 'skybot'),
        'tmp':  os.path.join(target_dir, 'tmp'),
        'weights':  os.path.join(target_dir, 'weights')
    }

    for _, path in paths.items():
        os.makedirs(path, exist_ok=True)

    return target_dir, paths




name = 'ssos'
version = '1.1.0'


GREETING = '''
       *       +      ssos pipeline %s
   +        x
        '             https://sso-pipeline.readthedocs.io
*   +     _.::'   |
        :_.'    - o - ssos --default : Copy default settings to CWD
|                 |   ssos --help    : Print detailed help
O -     *        '    ssos [-PARAMETER] [-VALUE] path/to/images
|
        _..:'         max.mahlke(at)cab.inta-csic.es
  +   (_.:

        ''' % version


'''
    Python implementations of filter functions to detect SSOs in SExtractor and SCAMP catalogs.

    Part of the SSO recovery pipeline.
'''

import collections

import numpy as np
import pandas as pd
from statsmodels import robust
from scipy import stats
from scipy.optimize import curve_fit


def detections(sources, settings):
    '''
        Remove sources based on number of detections

        input
        ------
        sources - pandas dataframe, table of source data
        detections - list, containing integers responding to number of detections

        return
        ------
        sources - pandas dataframe, sources data without the sources filtered in this step
    '''
    measurements = collections.Counter(sources['SOURCE_NUMBER'])
    sources_to_remove = {key for key, value in measurements.items()
                         if value in settings['DETECTIONS']}

    sources = sources[~sources['SOURCE_NUMBER'].isin(sources_to_remove)]
    return sources


def proper_motion(sources, settings):
    '''
    Remove sources based on the proper motion range and signal-to-noise

    input
    ------
    sources - pandas dataframe, SCAMP full catalog
    settings - dictionary, settings parameter:value pairs

    return
    ------
    sources - pandas dataframe, sources data without the sources filtered in this step
    '''

    pm_low = settings['PM_LOW']
    pm_up = settings['PM_UP']
    SN = settings['PM_SNR']

    # Filter sources for proper motion range
    sources = sources[(sources.PM > pm_low) & (sources.PM < pm_up)]

    # Filter by proper motion SN
    sources = sources[sources.PM / sources.PMERR > SN]
    return sources


def pixel(sources, settings):
    '''
    Remove sources based on pixel coordinates

    input
    ------
    sources - pandas dataframe, SCAMP full catalog
    settings - dictionary, contains the specified filter parameters

    return
    ------
    sources - pandas dataframe, sources data without the sources filtered in this step
    '''

    delta_pixel = settings['DELTA_PIXEL']

    # For all sources, check if the largest XWIN_IMAGE !and! YWIN_IMAGE differences
    # are smaller than delta_pixel.
    for source_number, group in sources.copy().groupby('SOURCE_NUMBER'):

        x_img = group['XWIN_IMAGE']
        y_img = group['YWIN_IMAGE']

        if x_img.max() - x_img.min() < delta_pixel:
            if y_img.max() - y_img.min() < delta_pixel:
                sources = sources[sources.SOURCE_NUMBER != source_number]

    return sources


def calc_r_squared(y_fit, y_data):
    '''
    Calculates the r-squared fit parameter to verify the linear motion

    input
    ------
    y_fit - list, values of fit function
    y_data - list, data points

    return
    ------
    r_squared - float, between 0 and 1
    '''

    residuals = y_data - y_fit
    ss_res = np.sum(residuals**2)
    ss_tot = np.sum((y_data-np.mean(y_data))**2)

    r_squared = 1 - (ss_res / ss_tot)
    return r_squared


def linear_function(x, m, n):
    return x*m + n


def motion_is_linear(epoch, ra, dec, r_squared):
    '''
    Motion fitting function

    input
    ------
    epoch - pd.Series, epoch of source
    ra - pd.Series, right ascension of source
    dec - pd.Series, declination of source
    r_squared - float, minimum r_squared for good fit

    return
    ------
    True if motion is linear else False
    '''

    for i, dimension in enumerate([ra, dec]):
        x = np.array(epoch)
        y = np.array(dimension)

        try:
            popt, pcov = curve_fit(linear_function, x, y)
        except RuntimeError:
            return False # If fit does not converge, discard this source

        fitted_r_squared = calc_r_squared(linear_function(x, *popt), y)

        if fitted_r_squared >= r_squared and i == 0:
            pass
        elif fitted_r_squared >= r_squared and i == 1:
            return True # motion is linear in both dimensions
        else:
            return False


def linear_motion(sources, settings):
    '''
        Remove sources based on linearity of motion and the Median Absolute Deviation

        input
        ------
        sources - pandas dataframe, table of source data
        settings - dict, 'parameter': value

        return
        ------
        sources - pandas dataframe, sources data without the sources filtered in this step
    '''

    r_squared = settings['R_SQU_M']
    outlier_threshold = settings['OUTLIER_THRESHOLD']
    min_detections = max(settings['DETECTIONS']) # lower limit of subgroup size
    sources_to_remove = []

    for source_number, group in sources.copy().groupby('SOURCE_NUMBER'):

        epoch = group['EPOCH']
        ra = group['ALPHA_J2000']
        dec = group['DELTA_J2000']

        if settings['IDENTIFY_OUTLIER']:
            # Detect outliers using the Median Absolute Deviation of the EPOCHS
            outlier_index = np.where(abs(np.diff(epoch)) > robust.mad(epoch) *
                                                           outlier_threshold)[0]
            if len(outlier_index)  == 0:  # no outlier in EPOCH. Check edges
                outlier_in_middle =  np.where(abs(np.diff(epoch[:-1])) > robust.mad(epoch[:-1]) *
                                                                         outlier_threshold)[0]
                if len(outlier_in_middle) == 0:
                    if not motion_is_linear(epoch, ra, dec, r_squared):
                        sources_to_remove.append(source_number)
                        continue

                else: # outlier found in middle (jump in epochs)
                    outlier_index = outlier_in_middle

            # Found outlier in EPOCH
            epoch_groups = np.split(epoch, outlier_index + 1)
            ra_groups    = np.split(ra,    outlier_index + 1)
            dec_groups   = np.split(dec,   outlier_index + 1)

            for i, epochs in enumerate(epoch_groups):
                if len(epochs) <= min_detections: # if there's not sufficient detections
                    sources.loc[epochs.index, 'FLAGS_SSOS'] +=1  # Add outlier flag

                else:
                    if not motion_is_linear(epochs, ra_groups[i], dec_groups[i], r_squared):
                        sources_to_remove.append(source_number)

            
        # FLAG_SSOS is only uneven if detection is outlier
        if all(sources[sources.SOURCE_NUMBER == source_number]['FLAGS_SSOS'] % 2 == 1):
            sources_to_remove.append(source_number)


        else:
            if not motion_is_linear(epoch, ra, dec, r_squared):
                sources_to_remove.append(source_number)

    sources = sources[~sources.SOURCE_NUMBER.isin(sources_to_remove)]
    return sources


def compare_std_and_sigma_of_average(y, sigma):
    ''' Compares weighted average uncertainty to standard deviation '''

    weights = 1 / sigma
    weighted_average_uncertainty = 1 / np.sqrt(np.sum(weights))
    standard_deviation_of_data = np.std(y)
    ratio = weighted_average_uncertainty / standard_deviation_of_data

    return ratio


def constant_trail(sources, settings):
    '''
    Remove sources based on constant trail size

    input
    ------
    sources - pd.DataFrame, source data
    settings - dict, 'parameter': value

    return
    ------
    sources - pandas dataframe, sources data without the sources filtered in this step
    '''

    ratio = settings['RATIO']

    for source_number, group in sources.copy().groupby('SOURCE_NUMBER'):

        epochs = group['EPOCH']
        awin = group['AWIN_IMAGE']
        bwin = group['BWIN_IMAGE']

        for i, dimension in enumerate([awin, bwin]):
            x = np.array(epochs)
            y = np.array(dimension)

            variance = group['ERRAWIN_IMAGE'] if i == 0 else group['ERRBWIN_IMAGE']
            sigma = np.array([np.sqrt(var) if var != 0 else np.sqrt(np.mean(variance))
                              for var in variance])

            fitted_ratio = compare_std_and_sigma_of_average(y, sigma)

            if fitted_ratio >= ratio and i == 0:
                pass
            elif fitted_ratio >= ratio and i == 1:
                pass
            else:
                sources = sources[sources.SOURCE_NUMBER != source_number]

    return sources


def trail_distribution(sources, settings):
    '''
    Remove sources based on distribution of size parameters

    input
    ------
    sources - pandas dataframe, table of source data
    settings - dict, 'parameter': value

    return
    ------
    sources - pandas dataframe, sources data without the sources filtered in this step
    '''

    sigma = settings['SIGMA']

    mu_a, std_a = np.mean(a), np.std(a)
    mu_b, std_b = np.mean(b), np.std(b)

    # Check if for any source any AWIN or BWIN value outside mu + sigma * std
    for source_number, group in sources.copy().groupby('SOURCE_NUMBER'):
        if any(group['AWIN_IMAGE'] > mu_a + sigma * std_a):
            sources = sources[sources.SOURCE_NUMBER != source_number]
        if any(group['BWIN_IMAGE'] > mu_b + sigma * std_b):
            sources = sources[sources.SOURCE_NUMBER != source_number]

    return sources


def star_catalog(sources, settings):
    '''
        Remove sources based on proximity to stars. Requires the HYG stellar
        catalog.

        input
        ------
        sources - pandas dataframe, table of source data
        settings - dict, 'parameter': value

        return
        ------
        sources - pandas dataframe, sources data without the sources filtered in this step
    '''

    distance = settings['DISTANCE'] / 3600 # convert distance to degree

    # -----
    # Open the HYG catalog
    stars = pd.read_csv(settings['HYGCAT'])
    stars['ra'] *= 15    # Convert RA coordinates in the HYG catalog from decimal hours to degrees

    # Remove stars which are outside region of interest, given by the minimum and maximum
    # RA and Dec values of the source candiates plus the distance of the stellar regions
    upper_ra, lower_ra = (sources['ALPHA_J2000'].max() + distance,
                          sources['ALPHA_J2000'].min() - distance)
    upper_dec, lower_dec = (sources['DELTA_J2000'].max() + distance,
                            sources['DELTA_J2000'].min() - distance)

    stars = stars[(stars['ra'] > lower_ra) &\
                  (stars['ra'] < upper_ra)]
    stars = stars[(stars['dec'] > lower_dec) &\
                  (stars['dec'] < upper_dec)]

    for source_number, group in sources.copy().groupby('SOURCE_NUMBER'):

            mean_ra = np.mean(group['ALPHA_J2000'])
            mean_dec = np.mean(group['DELTA_J2000'])

            try:
                min_distance_to_star = min(np.sqrt((stars['ra'] - mean_ra)**2 +
                                                   (stars['dec'] - mean_dec)**2).abs())

            except ValueError:  # reports "min() arg is an empty sequence"
                continue

            if min_distance_to_star < distance:
                sources = sources[sources.SOURCE_NUMBER != source_number]

    return sources



'''
    Extraction of cutouts of SSOs using SWarp,
    cross-matching with the SkyBoT database, and other analyses.

    Part of the SSO recovery pipeline.
'''

import os
import urllib.request

from astropy.io import fits
from astropy.io.votable import parse_single_table
from astropy.time import Time
import numpy as np
import pandas as pd


def extract_cutouts(sources, settings, log, paths, args):
    '''
        Extract cutouts of sources found in SCAMP catalogues using SWARP

        input
        ------
        sources - pd.DataFrame, table of source data
        settings - dict, dictionary of pipeline settings with PARAM:VALUE pairs
        log - logging object
        paths - dict, absolute paths of target directories
        args - command line argument parser

        return
        ------
        sources - pd.DataFrame, table of source data (unchanged)
    '''

    log.info('\nExtracting cutouts with SWARP..')

    image_dir    = args.fields[0]
    cutout_dir   = paths['cutouts']
    cutout_size  = settings['CUTOUT_SIZE']
    swarp_config = settings['SWARP_CONFIG']

    for index, row in sources.iterrows():

        image_file = os.path.join(image_dir, row['FILENAME_EXP']) + '[%i]' % row['SCI_EXTENSION']
        cutout_filename = os.path.join(cutout_dir, '{:.0f}_{:.0f}.fits'.format(row['SOURCE_NUMBER'],
                                                                               row['CATALOG_NUMBER']))

        if not args.swarp and os.path.isfile(cutout_filename):
            log.debug('Cutout %s exists, skipping..' % cutout_filename)
            continue

        swarp_args = {
                    'file': image_file,
                    'config': swarp_config,
                    'overwrite_params': {
                        'WEIGHTOUT_NAME': os.path.join(paths['tmp'], 'tmp_weight'),
                        'CENTER'        : '{:.5f},{:.5f}'.format(row['ALPHA_J2000'], row['DELTA_J2000']),
                        'IMAGEOUT_NAME' : '{:s}'.format(cutout_filename),
                        'IMAGE_SIZE'    : str(cutout_size),
                        'COPY_KEYWORDS' : 'OBJECT,RA,DEC,FILTER,DATE-OBS'},
                    }
        # ------
        # Build command and execute SWARP
        cmd = ' '.join(['swarp', swarp_args['file'], '-c', swarp_args['config']])
        for param, value in swarp_args['overwrite_params'].items():
            cmd += ' '.join([' -' + param, value])

        log.debug('Executing SWARP command:\n%s' % cmd)
        if log.level > 10:  # Shown SWARP warnings only when debugging
            cmd += ' >/dev/null 2>&1'
        os.system(cmd)

    log.info('\tDone.\n')
    return sources


def compute_aperture_magnitudes(sources, settings, log, paths, args):
    '''
    Executes SExtractor in dual image mode and computes fixed aperture magnitudes

    input
    ------
    sources - pd.DataFrame, table of source data
    settings - dict, dictionary of pipeline settings with PARAM:VALUE pairs
    log - logging object
    paths - dict, absolute paths of target directories
    args - command line argument parser

    return
    ------
    sources - pd.DataFrame, table of source data with new magnitudes
    '''

    log.info('Computing fixed aperture magnitudes with SExtractor..')

    cutout_dir = paths['cutouts']
    cutouts = [os.path.join(cutout_dir, cutout) for cutout in os.listdir(cutout_dir)
               if cutout.endswith('.fits')]
    reference_filter = settings['REFERENCE_FILTER'].split(',')

    for number, source in sources.groupby('SOURCE_NUMBER'):

        # Select detection image based on reference filter
        filter_found = False
        for filter_ in reference_filter:
            try:
                reference_detection = source.loc[source['FILTER_EXP'] == filter_].iloc[0]
                filter_found = True
            except IndexError:
                continue
            if filter_found:
                sources.loc[(sources.SOURCE_NUMBER == reference_detection.SOURCE_NUMBER) &
                            (sources.CATALOG_NUMBER == reference_detection.CATALOG_NUMBER), 'FLAGS_SSOS'] += 2
                break
        else:
            from core import PipelineSettingsException
            raise PipelineSettingsException('No detection of candidate %i found in reference filter %s.'\
                                            % (number, reference_filter))

        detection_image = os.path.join(cutout_dir, '_'.join([str(number),
                                                             str(reference_detection['CATALOG_NUMBER'])])
                                                             + '.fits')

        # ------
        # Call SExtractor in dual image mode
        for image in cutouts:

            cat = os.path.join(paths['cutouts'], os.path.basename(image.replace('.fits', '.cat')))

            if image == detection_image:
                detection_cat = cat

            sex_args = {
                'file': ','.join([detection_image, image]),
                'config': settings['SEX_CONFIG'],
                'overwrite_params': {
                    'PARAMETERS_NAME': settings['SEX_PARAMS'],
                    'CATALOG_NAME'   : cat,
                    'WEIGHT_TYPE'    : 'BACKGROUND',
                    'FILTER_NAME'    : settings['SEX_FILTER'],
                    'STARNNW_NAME'   : settings['SEX_NNW'],
                },
            }

            # ------
            # Exectue SExtractor
            cmd = ' '.join(['sex', sex_args['file'], '-c', sex_args['config']])
            for param, value in sex_args['overwrite_params'].items():
                cmd += ' '.join([' -' + param, value])
            os.system(cmd)

        # Identify the SSO in the detection cat to track it among the other ones
        cutout_size = settings['CUTOUT_SIZE']

        with fits.open(detection_cat) as detection_cat:

            # ------
            # Find the source closes to the center coordinate
            x_images = detection_cat[2].data.field('XWIN_IMAGE')
            y_images = detection_cat[2].data.field('YWIN_IMAGE')

            # Get 2d distances of sources to image center
            distances = [np.linalg.norm(np.array([cutout_size, cutout_size]) / 2 - np.array([x_img, y_img]))
                         for x_img, y_img in zip(x_images, y_images)]

            # Find index of source closest to center -> SSO candidate
            center_source_ind = np.where(distances == np.min(distances))[0]

        for ind, detection in source.iterrows():

            catalog = os.path.join(cutout_dir, '_'.join([str(detection.SOURCE_NUMBER),
                                                         str(detection.CATALOG_NUMBER)]) + '.cat')

            with fits.open(catalog) as cat:
                for prop in ['MAG', 'FLUX']:
                    sources.loc[ind, prop + '_CI'] = cat[2].data.field(prop + '_AUTO')[center_source_ind]
                    sources.loc[ind, prop + 'ERR_CI'] = cat[2].data.field(prop + 'ERR_AUTO')[center_source_ind]

    log.info('\tDone.\n')

    return sources


def _query_and_cross_match(group, target_dir, fov, obs_code, crossmatch_radius, log, args):
    '''
    Retrieves the SkyBoT query XML file for sources grouped by the DATE-OBS keyword

    input
    ------
    sources - pd.DataFrame, table of source data
    target_dir - str, absolute path to SkyBoT folder
    fov - str, value of FOV_DIMENSIONS parameter
    obs_code - str, value of OBSERVATORY_CODE parameter
    crossmatch_radius - float, crossmatch radius in degree
    log - logging object
    args - command line argument parser

    return
    -------
    group - pd.DataFrame group, sources grouped by DATE-OBS with SkyBoT data added
    '''


    # ------
    # Download SkyBoT query results to VOTABLE

    # Get mid-exposure epoch
    epoch = group['MID_EXP_MJD'].tolist()[0]
    epoch = Time(epoch, format='mjd')
    mid_exposure = epoch.isot

    ra_field = group['RA_EXP'].tolist()[0]
    dec_field = group['DEC_EXP'].tolist()[0]

    query_url = 'http://vo.imcce.fr/webservices/skybot/skybotconesearch_query.php?-ep=%s&-ra=%s&-dec=%s'\
                '&-bd=%s&-mime=votable&-output=basic&-loc=%s&-filter=0&-objFilter=111&&-from=AF&-top=' %\
                (mid_exposure, ra_field, dec_field, fov, obs_code)

    output_filename = 'skybot_%s_%s_%s.xml' % (ra_field, dec_field, mid_exposure)
    output_filename = os.path.join(target_dir, output_filename)

    if not args.skybot and os.path.isfile(output_filename):
            log.debug('\nAlready queried %s, skipping download..' % output_filename)
    else:
        urllib.request.urlretrieve(query_url, output_filename)

    # ------
    # Cross-match query results with sources from that epoch
    try:
        # Temporarily disable warnings as otherwise astropy.io.votable.tree
        # will spam the output with xml warnings
        import warnings
        warnings.filterwarnings('ignore')
        skybot_sources = parse_single_table(output_filename).array
        warnings.filterwarnings('default')
        skybot_sources = pd.DataFrame(np.ma.filled(skybot_sources))

    except IndexError:
        # No matches in SkyBoT database were found or error in query
        skybot_sources = pd.DataFrame({'number': [''],
                                       'name': [''],
                                       'magV': [0],
                                       '_raj2000': [0],
                                       '_decj2000': [0],
                                       'dracosdec': [0],
                                       'ddec': [0],
                                       'class': ['']})

    # Call actual cross-matching method
    group = group.apply(_cross_match, args=(skybot_sources, crossmatch_radius), axis=1)

    return group


def _cross_match(source, skybot_sources, crossmatch_radius):
    '''
        Cross-matching is applied to single rows in database.  Checks for matches
        and adds SkyBoT source parameters to database if a match is found.

        input
        ------
        source - pd.Series, single source detection parameters
        skybot_sources - pd.DataFrame, parameters of all SkyBoT sources in exposure
        crossmatch_radius - float, crossmatch radius in arcsec

        returns
        source - pd.Series, single source detection parameters with SkyBot parameters added
    '''

    # Find SkyBoT source that is closest in angular separation to KiDS source  # in degrees
    distances_to_source = abs(np.sqrt((skybot_sources['_raj2000'] - source['ALPHA_J2000'])**2 +\
                                      (skybot_sources['_decj2000'] - source['DELTA_J2000'])**2))
    min_distance_to_skybot_source = min(distances_to_source)

    if min_distance_to_skybot_source < crossmatch_radius:

        skybot_source_index = distances_to_source.argsort()[:1][0]
        skybot_match = skybot_sources.iloc[skybot_source_index]

        source['SKYBOT_NUMBER']  = skybot_match['num']
        source['SKYBOT_NAME']    = skybot_match['name']
        source['SKYBOT_MAG']     = skybot_match['magV']
        source['SKYBOT_ALPHA']   = skybot_match['_raj2000']
        source['SKYBOT_DELTA']   = skybot_match['_decj2000']
        source['SKYBOT_PMALPHA'] = skybot_match['dracosdec']
        source['SKYBOT_PMDELTA'] = skybot_match['ddec']
        source['SKYBOT_CLASS']   = skybot_match['class']

    return source


def crossmatch_skybot(sources, settings, log, paths, args):
    '''
    Query SkyBoT for each epoch and cross-match the expected SSOs
    with the sources in the database.  Requires working internet connection.

    input
    ------
    sources - pd.DataFrame, table of source data
    settings - dict, dictionary of pipeline settings with PARAM:VALUE pairs
    log - logging object
    paths - dict, absolute paths of target directories
    args - command line argument parser

    return
    ------
    sources - pd.DataFrame, table of source data with SkyBoT matches added
    '''
    log.info('\nCross-matching SSO candidates with SkyBoT..')


    target_dir = paths['skybot']
    fov        = settings['FOV_DIMENSIONS']
    obs_code   = settings['OBSERVATORY_CODE']
    crossmatch_radius = settings['CROSSMATCH_RADIUS'] / 3600 # convert from arcsecond to degrees


    # The default value for the SkyBoT columns are empty strings
    sources = pd.concat([sources, pd.DataFrame('', sources.index,
                        ['SKYBOT_NAME', 'SKYBOT_NUMBER', 'SKYBOT_CLASS', 'SKYBOT_MAG',
                         'SKYBOT_ALPHA', 'SKYBOT_DELTA', 'SKYBOT_PMALPHA', 'SKYBOT_PMDELTA'])], axis=1)

    # Cross-match for each mid-exposure epoch
    sources = sources.groupby('DATE-OBS_EXP').apply(_query_and_cross_match, target_dir=target_dir,
                                                    fov=fov, obs_code=obs_code, log=log, args=args,
                                                    crossmatch_radius=crossmatch_radius)

    log.info(' %i matches found.\n' % len(set(sources[sources['SKYBOT_NAME']!='']['SOURCE_NUMBER'])))
    return sources